/*
 * kgr_patch_bsc1102682
 *
 * Fix for CVE-2018-5390, bsc#1102682
 *
 *  Upstream commits:
 *  72cd43ba64fc ("tcp: free batches of packets in tcp_prune_ofo_queue()")
 *  f4a3313d8e2c ("tcp: avoid collapses in tcp_prune_queue() if possible")
 *  3d4bf93ac120 ("tcp: detect malicious patterns in tcp_collapse_ofo_queue()")
 *  8541b21e781a ("tcp: call tcp_drop() from tcp_data_queue_ofo()")
 *  58152ecbbcc6 ("tcp: add tcp_ooo_try_coalesce() helper")
 *
 *  SLE12 commit:
 *  none yet
 *
 *  SLE12-SP1 commit
 *  none yet
 *
 *  SLE12-SP2 and -SP3 commits:
 *  575a71e9fb4e4bc5955b3a27bc139c9c10023dae
 *  a5ae92abdf62a5d34d9388496b3783ba5b23b8f0
 *  401f16adaae56c561591ce4b0cbb8a6c5ec83f3c ("Linux 4.4.149")
 *
 *  SLE15 commits:
 *  ec3d22a4c081f5988fd20d088c68201f5196a5e7
 *  fe0ec2a50973fb4bf2c2cecd6f1efd90b5804739
 *  8fc8cfda8b5c1af4e3e4b121737acb9f155d3b0c
 *  6ee46ea3308017deb1f444dc2e1225c74e42af90
 *  ce20133201e35e2e1d3030112b8a257a51b0951f
 *  b01b19a87acb29b2e36b5a79dcbf140b7fb5c639
 *
 *
 *  Copyright (c) 2018 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <net/tcp.h>
#include "kgr_patch_bsc1102682.h"
#include "kallsyms_relocs.h"

static void
(*kgr_tcp_collapse)(struct sock *sk, struct sk_buff_head *list,
		    struct sk_buff *head, struct sk_buff *tail,
		    u32 start, u32 end);
static bool (*kgr_tcp_prune_ofo_queue)(struct sock *sk);

static struct kgr_kallsyms_reloc kgr_funcs[] = {
	{"tcp_collapse", (void *)&kgr_tcp_collapse},
	{"tcp_prune_ofo_queue", (void *)&kgr_tcp_prune_ofo_queue},
};



/* from net/ipv4/tcp_input.c */
/* inlined */
static void kgr_tcp_clamp_window(struct sock *sk)
{
	struct tcp_sock *tp = tcp_sk(sk);
	struct inet_connection_sock *icsk = inet_csk(sk);

	icsk->icsk_ack.quick = 0;

	if (sk->sk_rcvbuf < sysctl_tcp_rmem[2] &&
	    !(sk->sk_userlocks & SOCK_RCVBUF_LOCK) &&
	    !tcp_under_memory_pressure(sk) &&
	    sk_memory_allocated(sk) < sk_prot_mem_limits(sk, 0)) {
		sk->sk_rcvbuf = min(atomic_read(&sk->sk_rmem_alloc),
				    sysctl_tcp_rmem[2]);
	}
	if (atomic_read(&sk->sk_rmem_alloc) > sk->sk_rcvbuf)
		tp->rcv_ssthresh = min(tp->window_clamp, 2U * tp->advmss);
}



/* patched, inlined into also patched tcp_prune_queue() */
static void kgr_tcp_collapse_ofo_queue(struct sock *sk)
{
	struct tcp_sock *tp = tcp_sk(sk);
	/*
	 * Fix CVE-2018-5390
	 *  +1 line
	 */
	u32 range_truesize, sum_tiny = 0;
	struct sk_buff *skb = skb_peek(&tp->out_of_order_queue);
	struct sk_buff *head;
	u32 start, end;

	if (!skb)
		return;

	start = TCP_SKB_CB(skb)->seq;
	end = TCP_SKB_CB(skb)->end_seq;
	/*
	 * Fix CVE-2018-5390
	 *  +1 line
	 */
	range_truesize = skb->truesize;
	head = skb;

	for (;;) {
		struct sk_buff *next = NULL;

		if (!skb_queue_is_last(&tp->out_of_order_queue, skb))
			next = skb_queue_next(&tp->out_of_order_queue, skb);
		skb = next;

		/* Segment is terminated when we see gap or when
		 * we are at the end of all the queue. */
		if (!skb ||
		    after(TCP_SKB_CB(skb)->seq, end) ||
		    before(TCP_SKB_CB(skb)->end_seq, start)) {
			/*
			 * Fix CVE-2018-5390
			 *  -2 lines, + 11 lines
			 */
			/* Do not attempt collapsing tiny skbs */
			if (range_truesize != head->truesize ||
				end - start >= SKB_WITH_OVERHEAD(SK_MEM_QUANTUM)) {
				kgr_tcp_collapse(sk, &tp->out_of_order_queue,
					head, skb, start, end);
			} else {
				sum_tiny += range_truesize;
				if (sum_tiny > sk->sk_rcvbuf >> 3)
					return;
			}

			head = skb;
			if (!skb)
				break;
			/* Start new segment */
			start = TCP_SKB_CB(skb)->seq;
			end = TCP_SKB_CB(skb)->end_seq;
			/*
			 * Fix CVE-2018-5390
			 *  +1 line
			 */
			range_truesize = skb->truesize;
		} else {
			/*
			 * Fix CVE-2018-5390
			 *  +1 line
			 */
			range_truesize += skb->truesize;
			if (before(TCP_SKB_CB(skb)->seq, start))
				start = TCP_SKB_CB(skb)->seq;
			if (after(TCP_SKB_CB(skb)->end_seq, end))
				end = TCP_SKB_CB(skb)->end_seq;
		}
	}
}

/* patched, inlined */
static int kgr_tcp_prune_queue(struct sock *sk)
{
	struct tcp_sock *tp = tcp_sk(sk);

	SOCK_DEBUG(sk, "prune_queue: c=%x\n", tp->copied_seq);

	NET_INC_STATS_BH(sock_net(sk), LINUX_MIB_PRUNECALLED);

	if (atomic_read(&sk->sk_rmem_alloc) >= sk->sk_rcvbuf)
		kgr_tcp_clamp_window(sk);
	else if (tcp_under_memory_pressure(sk))
		tp->rcv_ssthresh = min(tp->rcv_ssthresh, 4U * tp->advmss);

	/*
	 * Fix CVE-2018-5390
	 *  +3 lines
	 */
	if (atomic_read(&sk->sk_rmem_alloc) <= sk->sk_rcvbuf)
		return 0;

	kgr_tcp_collapse_ofo_queue(sk);
	if (!skb_queue_empty(&sk->sk_receive_queue))
		kgr_tcp_collapse(sk, &sk->sk_receive_queue,
				 skb_peek(&sk->sk_receive_queue),
				 NULL,
				 tp->copied_seq, tp->rcv_nxt);
	sk_mem_reclaim(sk);

	if (atomic_read(&sk->sk_rmem_alloc) <= sk->sk_rcvbuf)
		return 0;

	/* Collapsing did not help, destructive actions follow.
	 * This must not ever occur. */

	kgr_tcp_prune_ofo_queue(sk);

	if (atomic_read(&sk->sk_rmem_alloc) <= sk->sk_rcvbuf)
		return 0;

	/* If we are really being abused, tell the caller to silently
	 * drop receive data on the floor.  It will get retransmitted
	 * and hopefully then we'll have sufficient space.
	 */
	NET_INC_STATS_BH(sock_net(sk), LINUX_MIB_RCVPRUNED);

	/* Massive buffer overcommit. */
	tp->pred_flags = 0;
	return -1;
}

/* patched, calls inlined tcp_prune_queue() */
int kgr_tcp_try_rmem_schedule(struct sock *sk, struct sk_buff *skb,
			      unsigned int size)
{
	if (atomic_read(&sk->sk_rmem_alloc) > sk->sk_rcvbuf ||
	    !sk_rmem_schedule(sk, skb, size)) {

		if (kgr_tcp_prune_queue(sk) < 0)
			return -1;

		if (!sk_rmem_schedule(sk, skb, size)) {
			if (!kgr_tcp_prune_ofo_queue(sk))
				return -1;

			if (!sk_rmem_schedule(sk, skb, size))
				return -1;
		}
	}
	return 0;
}



int kgr_patch_bsc1102682_init(void)
{
	return __kgr_resolve_kallsyms_relocs(kgr_funcs, ARRAY_SIZE(kgr_funcs));
}
